#include <SoftwareSerial.h>
SoftwareSerial SSerial(2,3); // RX, TX

//#define DBGSerial     1
//#define DBGab 

static const uint16_t c_KeypadMapping[] = {
  1, 4, 7, 0,
  2, 5, 8, 0xf,
  3, 6, 9, 0xe, 
  0xa, 0xb, 0xc, 0xd};


  
#define FIRSTANALOG            1
#define NUMANALOGS             3

#define FIRST_KEYPAD_ROWPIN    9  // 0-7 used for Keypad
#define FIRST_KEYPAD_COLPIN    4
#define NROW                   4
#define NCOL                   4 

#define MODE_NORMAL            1
#define MODE_CALIBRATE         2

#define XBEE_API_PH_SIZE    1
#define XBEE_TRANS_DATA        0x03  // Data Packet from Transmitter to Robot*
#define XBEE_ENTER_SSC_MODE     0x05  // The controller is letting robot know to enter SSC echo mode
#define XBEE_RECV_DISP_VAL      0x84  // Display a value on line 2
// If <cbExtra> is  0 then we will display the number contained in <SerialNumber> 
// If not zero, then it is a count of bytes in a string to display.
#define XBEE_RECV_DISP_STR      0x85  // Display a string value on line 2
#define XBEE_PLAY_SOUND       0x86  // Will make sounds on the remote...
//  <cbExtra> - 2 bytes per sound: Duration <0-255>, Sound: <Freq/25> to make fit in byte...
#define XBEE_SSC_MODE_EXITED    0x87  // a message sent back to the controller when
// it has left SSC-mode.
#define XBEE_SEND_SN_NI_DATA    0x88  // Response for REQ_SN_NI - will return
// 4 bytes - SNH
// 4 bytes - SNL
// up to 20 bytes(probably 14) for NI
#define XBEE_RECV_DISP_VAL0   0x89  // Display a 2nd value on line 2 - Col 0 on mine
#define XBEE_RECV_DISP_VAL1   0x8A  // Display a 2nd value on line 2
#define XBEE_RECV_DISP_VAL2   0x8B  // Display a value on line 2  - Cal 



enum {
  PKT_BTNLOW=0,     // Low Buttons 0-7
  PKT_BTNHI,         // High buttons 8-F
  PKT_RJOYLR,        // Right Joystick Up/Down
  PKT_RJOYUD,        // Right joystick left/Right
  PKT_LJOYLR,  // Left joystick Left/Right
  PKT_LJOYUD, // Left joystick Up/Down
  PKT_RSLIDER,  // right slider
  PKT_LSLIDER,  // Left slider
  PKT_RPOT,          // Right Pot
  PKT_LPOT,          // Left Pot
  PKT_MSLIDER,       // Middle Slider
  PKT_BTNS2          // Extra buttons like on top of joystick
};


// Main data packet            
typedef union {
  byte    ab[12];
  struct  {    
    uint16_t  wButtons;         // the 16 buttons
    byte  bRJoyLR;          // right joystick X (LR)
    byte  bRJoyUD;          //        ..Y (UD)
    byte  bLJoyLR;          // Left Joystick  X (LR)
    byte  bLJoyUD;          //          Y (UD)
    byte  bRSlider;         // Right Slider
    byte  bLSlider;         // Left Slider
    // Added values since first version                    
    byte    bRPot;                                          // top Pot on right joystick
    byte    bLPot;                                          // ... Left joystick
    byte  bMSlider;         // Middle Slider
    byte    bButtons2;                                      // Extra buttons on new DIY...
  } 
  s;
} 
DIYPACKET;

typedef DIYPACKET *PDIYPACKET;

typedef unsigned long ulong;

// Now define some static state stuff
typedef struct _diystate
{
  // state information

  // More internal to function...
  byte            bPacketNum;
  byte            bAPIPacket[33];        // Api packet
  uint16_t    wAPIDL;                // current destination.
  uint16_t        wDataPacketDeltaMS;    // How long should we wait between data packets   

  // Other information, could make static to file...
  ulong           ulLastPacket;

  uint16_t        wXBeeMY;                              // Our My value
  uint16_t        wXBeeDL;                              // Current DL

} 
DIYSTATE;

DIYSTATE g_diystate;




 
//=============================================================================
// Global Variables
//=============================================================================
uint16_t    wKeypadPrev;

// Mode information
byte        g_bMode;                                       // What mode are we in...
boolean     g_fNewMode;                                    // Have we just entered this mode?

uint16_t    g_aawRawAnalog[8][NUMANALOGS];                 // Keep 8 raw values for each analog input
uint16_t    g_awRawSums[NUMANALOGS];                       // sum of the raw values.

uint16_t    g_awAnalogMids[NUMANALOGS];                // This is our observed center points at init time. 
int         g_iRaw;                   // which raw value are we going to read into


static const byte g_aiRawToDIYP[] = {
  PKT_RJOYLR, PKT_RJOYUD, PKT_LJOYLR, PKT_LJOYUD, PKT_RSLIDER, PKT_LSLIDER,  
  PKT_RPOT, PKT_LPOT, PKT_MSLIDER};

static const boolean  g_afAnalogUseMids[] = {
  1,1,1,1,0,0,1,1, 0};       // Do we use the mid points when we normalize
static const uint16_t g_awAnalogMins[] = {
  0*8,0*8,0*8,0*8,0*8,0*8, 0*8, 0*8, 0*8};  // Will have a calibrate later where we read these in (if necessary)
static const uint16_t g_awAnalogMaxs[] = {
  1023*8,1023*8,1023*8,1023*8,1023*8,1023*8, 1023*8, 1023*8, 1023*8}; // dito

  

DIYPACKET diyp;        // Our data packet we will send out...
DIYPACKET diypSave;

uint8_t abT[10]; 


void setup() {
  SSerial.begin(57600);
  Serial.begin(115200);

  g_diystate.bPacketNum = 0;
  g_diystate.wDataPacketDeltaMS = 26;
  Serial.setTimeout(2000);
  
  InitKeypad(); 
  InitJoysticks();
  g_bMode = MODE_NORMAL;


  delay(2000);
}


boolean bInit=false;
void loop() {
  char ch;
  uint16_t val1;
  
  
  diyp.s.wButtons = ReadKeypad(&ch);
  ReadJoysticks();


switch (g_bMode) {
  case MODE_NORMAL:
    CheckAndTransmitDataPacket(&diyp);

    if (diyp.s.wButtons != wKeypadPrev) {
      //Serial.print(ch); // g_display.DisplayKey(ch);
    }
   
   
   //val1 =   (abT[0]<<8 ) | abT[1];  bytes to int 

    CheckAndTransmitDataPacket(&diyp);
#ifdef DBGSerial
    if (memcmp(&diyp, &diypSave, sizeof(diyp))) {
        memcpy(&diypSave, &diyp, sizeof(diyp)); 
        SSerial.print(diyp.s.wButtons,HEX);
        SSerial.print(":");
        SSerial.print(diyp.s.bRJoyLR,DEC);
        SSerial.print(":");
        SSerial.print(diyp.s.bRJoyUD,DEC);
        SSerial.print(":");
        SSerial.print(diyp.s.bLJoyLR,DEC);
        SSerial.print(":");
        SSerial.print(diyp.s.bLJoyUD,DEC);
        SSerial.print(":");
        SSerial.print(diyp.s.bRPot,DEC);
        SSerial.print(":");
        SSerial.print(diyp.s.bLPot,DEC);
        SSerial.println();
        }
#endif
    break;
  case MODE_CALIBRATE:
    CalibrateMode();
    break;    
  } // End of our mode case statement...

// Only at end of pass update some previous states as function may need to know if anything changed...
  wKeypadPrev = diyp.s.wButtons;
  //memcpy(&abT, &diyp, sizeof(diyp));
  memcpy(&diypSave, &diyp, sizeof(diyp));      
    

}

//==============================================================================
// CalibrateMode: Go through and Calibrate all of the joysticks from their minimum
//     to maximum values to allow us to better output a standardized range...
//==============================================================================
void CalibrateMode(void) {
}

//=============================================================================
// InitJoysticks: Initialize our Joysticks and other analog devices
//=============================================================================
void InitJoysticks(void) {          // Initialize the joysticks.
  byte i;
  uint16_t w;
  // Later may need to read in Min/Max values for each of the Anlog inputs from EEPROM.

  // We need to prefill our array of raw items and sums.  The first read for each analog
  // may be trash, but that will get removed on our first real read.
  // First do our Real Analogs...
  for (i=0; i<NUMANALOGS; i++) {
    g_awRawSums[i] = 0;        // initialize to zero
    w = analogRead(FIRSTANALOG+i);    // first read is not valid
    for (g_iRaw=0; g_iRaw < 8; g_iRaw++) {
      g_aawRawAnalog[g_iRaw][i] = analogRead(FIRSTANALOG+i);
      g_awRawSums[i] += g_aawRawAnalog[g_iRaw][i];
      delay(1);
    }
    // Save away these sums as our mid points...
    g_awAnalogMids[i] = g_awRawSums[i];
  }

}



//=============================================================================
// ReadJoysticks: 
//
//=============================================================================
uint16_t g_awAnalogsPrev[NUMANALOGS];

void ReadJoysticks() {          // Initialize the joysticks.
  byte i;
  boolean fChanged = false;

  // Calculate which index of our raw array we should now reuse.
  g_iRaw = (g_iRaw+1) & 0x7;

  // Now lets read in the next analog reading and smooth the values
  // Will separate out Normal analogs from MUX.  
  for (i=0; i<NUMANALOGS; i++) {
    uint16_t wAnalog = analogRead(FIRSTANALOG+i);
    ReadCalibrateAndNormalizeAnalog(g_iRaw, i, wAnalog);
    if (wAnalog != g_awAnalogsPrev[i]) {
      g_awAnalogsPrev[i] = wAnalog;
      fChanged = true;
    }
  }
/*#ifdef DBGSerial  
  if (fChanged) {
    Serial.print("A: ");
    for (i=0; i< NUMANALOGS; i++) {
      Serial.print(g_awAnalogsPrev[i], DEC);
      Serial.print(" ");
    }
    Serial.println();
  }
#endif
*/


}

//=============================================================================
// Calibrate and Normalize AnalogValue
//=============================================================================
void ReadCalibrateAndNormalizeAnalog(byte iRaw, byte i, uint16_t wAnalog) {
  g_awRawSums[i] -= g_aawRawAnalog[iRaw][i];        // remove the value we overwrite from sum
  g_aawRawAnalog[iRaw][i] = wAnalog;
  g_awRawSums[i] += wAnalog;        // Add the new value in
  // Lets calculate our calibrated values from 0-255 whith 128 as center point
  if (g_awRawSums[i] <= g_awAnalogMins[i])
    diyp.ab[g_aiRawToDIYP[i]] = 0;                // Take care of out of range low
  else if (g_awRawSums[i] >= g_awAnalogMaxs[i])
    diyp.ab[g_aiRawToDIYP[i]] = 255;              // out of range high
  else if (!g_afAnalogUseMids[i])
    diyp.ab[g_aiRawToDIYP[i]] = map(g_awRawSums[i], g_awAnalogMins[i], g_awAnalogMaxs[i], 0, 255);
  else if (g_awRawSums[i] <= g_awAnalogMids[i])
    diyp.ab[g_aiRawToDIYP[i]] = map(g_awRawSums[i], g_awAnalogMins[i], g_awAnalogMids[i], 0, 128);
  else
    diyp.ab[g_aiRawToDIYP[i]] = map(g_awRawSums[i], g_awAnalogMids[i], g_awAnalogMaxs[i], 128, 255);

  // If in normal mode make sure we transmit as often as possible...
  if (g_bMode==MODE_NORMAL)
    CheckAndTransmitDataPacket(&diyp);



#ifdef DBGab
for (i=0; i< NUMANALOGS; i++) {
      SSerial.print(diyp.ab[g_aiRawToDIYP[i]], DEC);
      SSerial.print(" ");
    }
    SSerial.println();
#endif

}


//=============================================================================
// InitKeypad: Lets setup the IO pins associated with the keypad
//=============================================================================

void InitKeypad(void) {
  // Assume first four pins are the rows and the next four pins are the cols.
  byte bPin;

  // Init Rows
  for (bPin = FIRST_KEYPAD_ROWPIN; bPin < FIRST_KEYPAD_ROWPIN+4; bPin++) {
    pinMode(bPin, INPUT);        // set all of the pins to input
  }

  // Init Cols
  for (bPin = FIRST_KEYPAD_COLPIN; bPin < FIRST_KEYPAD_COLPIN+4; bPin++) {
    pinMode(bPin, INPUT_PULLUP);        // set all of the pins to input
  }
}        

//=============================================================================
// ReadKeypad: Now lets read the keypad and see if any buttons are pressed
//=============================================================================
uint16_t ReadKeypad(char *pch) {
  uint16_t wRet = 0;
  byte bRow;
  byte bCol;
  byte bBit;

  *pch = ' ';    // assume a blank
  for (bRow = FIRST_KEYPAD_ROWPIN; bRow < FIRST_KEYPAD_ROWPIN+NROW; bRow++) { //4
    // Set the row to low
    pinMode(bRow, OUTPUT);
    digitalWrite(bRow, LOW);

    for (bCol = FIRST_KEYPAD_COLPIN; bCol <  FIRST_KEYPAD_COLPIN+NCOL; bCol++) { //9
      if (digitalRead(bCol) == LOW) {
        bBit =  c_KeypadMapping[(bRow-FIRST_KEYPAD_ROWPIN)*4 + bCol-(FIRST_KEYPAD_COLPIN)];
        wRet |= 1 << bBit;
        if (bBit <= 9){
          *pch = '0'+ bBit;
        }
        else{
          *pch = 'A'+ bBit-10;
        }
      
      }
    }
    // restore the IO pin to be an input...
    pinMode(bRow, INPUT_PULLUP);    // 
  }
  
  return wRet;
}

//==============================================================================
// [CheckAndTransmitDataPacket] function
// 
// This function will output a packet of data over the XBee to the receiving robot. 
//==============================================================================

byte g_bTxStatusLast = 99;

void CheckAndTransmitDataPacket(PDIYPACKET pdiyp) {
  byte bDataOffset;
  byte bPacketType;
  byte cbPacket;
  unsigned long ulCurTime = millis();

  // First see if it is time for us to send out the next data packet:
  if ((ulCurTime - g_diystate.ulLastPacket) > g_diystate.wDataPacketDeltaMS) {
    SendXBeePacket(XBEE_TRANS_DATA, sizeof(DIYPACKET), (byte*)pdiyp);    // Ok we dumped the data to the the output
    g_diystate.ulLastPacket = ulCurTime;
  }


  while (cbPacket = APIRecvPacket(0)) {
    // We received an XBee Packet, See what type it is.
    // first see if it is a RX 16 bit or 64 bit packet?
    if (g_diystate.bAPIPacket[0] == 0x81)
      bDataOffset = 5;     // 16 bit address sent, so there is 5 bytes of packet header before our data
    else if (g_diystate.bAPIPacket[0] == 0x80) 
      bDataOffset = 11;    // 64 bit address so our data starts at offset 11...
    else if (g_diystate.bAPIPacket[0] == 0x89) {
      if (g_diystate.bAPIPacket[2] != g_bTxStatusLast) {
        g_bTxStatusLast = g_diystate.bAPIPacket[2];

      }
      continue; // this is an A TX Status or API status  message - May check status and maybe update something?
    }
    else if (g_diystate.bAPIPacket[0] == 0x88)
      continue; // API status  message - May check status and maybe update something?
    else {

      break;    // lets just bail from this loop
    }

    cbPacket -= (bDataOffset + XBEE_API_PH_SIZE); // This is the extra data size
    bPacketType = g_diystate.bAPIPacket[bDataOffset + 0];

    if (bPacketType == XBEE_RECV_DISP_VAL )  {
      // We will handle word values here
      if (cbPacket == 2)
//        g_display.DisplayRemoteValue(2, (g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE] << 8) + g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE+1]);
      g_diystate.ulLastPacket = millis();
    }

    else if (bPacketType == XBEE_RECV_DISP_STR )  {
            
      if ((cbPacket > 0) && (cbPacket <= 16)) 
        g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE+cbPacket] = 0;    // zero terminate string.
//      g_display.DisplayStatus(0, (char*)&g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE]);
      g_diystate.ulLastPacket = millis();
    }   //

    else if ((bPacketType >= XBEE_RECV_DISP_VAL0 ) && (bPacketType <= XBEE_RECV_DISP_VAL2 )) {
      // We will handle word values here
      if (cbPacket == 2) 
//        g_display.DisplayRemoteValue(bPacketType-XBEE_RECV_DISP_VAL0,
//        (g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE] << 8) + g_diystate.bAPIPacket[bDataOffset+XBEE_API_PH_SIZE+1]);
      g_diystate.ulLastPacket = millis();
    }
    // Packet to play a sound - BUGBUG - not checking checksum...
    
    else {        // Unknown packet

     delay(5);  // do simple way.
      ClearXBeeInputBuffer(); // try to clear everything else out.
      break;    // and get out of this loop;
    }
  }
}

//////////////////////////////////////////////////////////////////////////////
//==============================================================================
// [APIRecvPacket - try to receive a packet from the XBee. 
//        - Will return true if it receives something, else false
//        - Pass in buffer to receive packet.  Assumed it is big enough...
//        - pass in timeout if zero will return if no data...
//        
//==============================================================================
byte APIRecvPacket(ulong Timeout)
{
  byte cbRead;
  byte abT[3];
  byte bChksum;
  int i;

  short wPacketLen;
  //  First see if the user wants us to wait for input or not
  //    hserstat HSERSTAT_INPUT_EMPTY, _TP_Timeout            // if no input available quickly jump out.
  if (Timeout == 0) 
  {
    if (!Serial.available())
      return 0;        // nothing waiting for us...
    Timeout = 100;    // .1 second?

  }

  Serial.setTimeout(Timeout);

  // Now lets try to read in the data from the xbee...
  // first read in the delimter and packet length
  // We now do this in two steps.  The first to try to resync if the first character
  // is not the proper delimiter...

  do {    
    cbRead = Serial.readBytes(abT, 1);
    if (cbRead == 0)
      return 0;
  } 
  while (abT[0] != 0x7e);

  cbRead = Serial.readBytes(abT, 2);
  if (cbRead != 2)
    return 0;                // did not read in full header or the header was not correct.

  wPacketLen = (abT[0] << 8) + abT[1];
  if (wPacketLen >= sizeof(g_diystate.bAPIPacket)) {
      
    ClearXBeeInputBuffer();  // maybe clear everything else out.
    return 0;  // Packet bigger than expected maybe bytes lost...
  }

  // Now lets try to read in the packet
  cbRead = Serial.readBytes(g_diystate.bAPIPacket, wPacketLen+1);


  // Now lets verify the checksum.
  bChksum = 0;
  for (i = 0; i < wPacketLen; i++)
    bChksum = bChksum + g_diystate.bAPIPacket[i];             // Add that byte to the buffer...


  if (g_diystate.bAPIPacket[wPacketLen] != (0xff - bChksum)) {
   
    return 0;                // checksum was off
  }
  return wPacketLen;    // return the packet length as the caller may need to know this...
}

//==============================================================================
// [ClearXBeeInputBuffer] - This simple helper function will clear out the input
//                        buffer from the XBEE
//==============================================================================

void ClearXBeeInputBuffer(void)
{
  while (Serial.read() != -1)
    ;    // just loop as long as we receive something...
}


//==============================================================================
// [SendXBeePacket] - Simple helper function to send the 4 byte packet header
//     plus the extra data if any
//      gosub SendXBeePacket[bPacketType, cbExtra, pExtra]
//==============================================================================
void SendXBeePacket(byte bPHType, byte cbExtra, byte *pbExtra)
{
  // Tell system to now output to the xbee
  byte abPH[9];
  byte *pbT;
  byte bChkSum;
  int i;

  // We need to setup the xbee Packet
  abPH[0]=0x7e;                        // Command prefix
  abPH[1]=0;                            // msb of size
  abPH[2]=cbExtra+XBEE_API_PH_SIZE + 5;    // size LSB
  abPH[3]=1;                             // Send to 16 bit address.

  g_diystate.bPacketNum = g_diystate.bPacketNum + 1;
  if (g_diystate.bPacketNum == 0)
    g_diystate.bPacketNum = 1;        // Don't pass 1 as this says no ack
  abPH[4]=g_diystate.bPacketNum;        // frame number
  abPH[5]=g_diystate.wAPIDL >> 8;        // Our current destination MSB/LSB
  abPH[6]=g_diystate.wAPIDL & 0xff;
  abPH[7]=0;                            // No Options

  abPH[8]=bPHType;

  // Now compute the initial part of the checksum
  bChkSum = 0;
  for (i=3;i <= 8; i++)
    bChkSum += abPH[i];

  // loop through the extra bytes in the exta to build the checksum;
  pbT = pbExtra;
  for (i=0; i < cbExtra; i++)
    bChkSum += *pbT++;                // add each byte to the checksum

  // Ok lets output the fixed part
  WriteToXBee(abPH,9); 

  // Ok lets write the extra bytes if any to the xbee
  if (cbExtra)
    WriteToXBee(pbExtra, cbExtra);

  // Last write out the checksum
  bChkSum = 0xff - bChkSum;
  WriteToXBee(&bChkSum, 1);


}



//=============================================================================
// void WriteToXBee - output a buffer to the XBEE.  This is used to abstract
//            away the actual function needed to talk to the XBEE
//=============================================================================
void WriteToXBee(byte *pb, byte cb)
{
  Serial.write(pb,cb);
  
 /* for (int i=0;i<=cb;i++){
  SSerial.print(pb[i],DEC);SSerial.print(" ");}
  SSerial.println();
  */
}
